-- --8<--8<--8<--8<--
--
-- Copyright (C) 2011 Smithsonian Astrophysical Observatory
--
-- This file is part of saotrace.baffles
--
-- saotrace.baffles is free software: you can redistribute it and/or modify
-- it under the terms of the GNU General Public License as published by
-- the Free Software Foundation, either version 3 of the License, or (at
-- your option) any later version.
--
-- This program is distributed in the hope that it will be useful,
-- but WITHOUT ANY WARRANTY; without even the implied warranty of
-- MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
-- GNU General Public License for more details.
--
-- You should have received a copy of the GNU General Public License
-- along with this program.  If not, see <http://www.gnu.org/licenses/>.
--
-- -->8-->8-->8-->8--

local seq = require('pl.seq')

local assert = assert
local string = require('string')
local error = error

local geometry = require( 'saotrace.geometry' )
local rdb = require('rdb')
local tables = require( 'saotrace.suplib.tables' )

local vobj = require( 'validate.args' ):new()
vobj:setopts{ named = true}

module( ... )

vspec = {

   shell  = { type = 'posnum' },
   optic  = { type = 'string' }, -- e.g. p, s for primary and secondary
   cols   = { vtable = {
		 shell = { type = 'string', default = 'shell' },
		 optic = { type = 'string', default = 'optic' },
		 rthick = { type = 'string', default = 'rthick' },
		 x = { type = 'string' },
		 y = { type = 'string' },
		 z = { type = 'string' },
		 r = { type = 'string' },
	      },
	   },
   db     = { type = 'string' },
   inner_disk = {
      optional = true,
      default_is_nil = true,
      vtable = { x = { type = 'number', default = 0 },
		 y = { type = 'number', default = 0 },
		 z = { type = 'number' },
		 r = { type = 'posnum' },
	      },
   },
}


--- create an ideal baffle from the geometry database
function create( ... )


   local ok, args = vobj:validate( vspec, ... )
   assert( ok, args )

   -- select the previous, current, or next shells
   local shell = args.shell
   local optic = args.optic


   local iter, vars = geometry.read{ file = args.db,
				     mode = 'iterator' }

   -- rename the fields to what we want them to be
   local extract = function(x) return tables.extract(x, nil, args.cols) end

   -- only grab those for the specified optic type
   local filter = function(x) return x.optic == optic end

   -- sort them in reverse radii
   local sort   = function(e1, e2) return e1.r > e2.r end

   -- the full iterator; note the last() at the end which couples neighboring
   -- entries
   local entries = seq(iter):map(extract):filter( filter ):sort( sort ):last()

   local inner, outer, prev

   for current, last in entries do

      if current.shell == shell then

	 prev = last
	 outer = current

      elseif last.shell == shell then

	 -- the last() sequence method tosses the first element in the sequence
	 -- so if shell is the first in the sequence it'll be the last element
	 -- in the first element of entries
	 outer = last

	 inner = current

      end

   end

   -- these will be assigned to the inner and outer aperture info

   -- make sure we found a match
   if not outer then

      error( string.format( "%s: unable to find mirror '%s %d'",
			    args.db, optic, shell ) )

      -- no smaller mirror
   elseif not inner then

      -- if there's only one mirror then inner_disk must be specified
      if not prev  and not args.inner_disk then
	 error( "only one mirror matched and inner_disk was not specified")
      end

      -- if inner_disk is specified use that
      if args.inner_disk then

	 inner = args.inner_disk

	 -- not specified, use the distance to the next outermost mirror
      else

	 inner = tables.copy( outer )
	 inner.r = outer.r - ( prev.r - (outer.r + outer.rthick ) )

      end

      -- next smaller was specified
   else

      inner.r = inner.r + inner.rthick

   end

   return inner, outer

end
